///////////////////////////////////////////////////////////////////////
// implementation of the CIApp class
// last updated: 14apr2004
// Copyright(C) 2004 Micronet Co. All Rights Reserved.

#include <iostream>
#include "rtpp.h"


///////////////////////////////////////////////////////////////////////
// pointer to the one and only application object

CIApp * CIApp::		m_pApp			= {(CIApp*)NULL};


///////////////////////////////////////////////////////////////////////
// constructor

CIApp::				CIApp(
	char *				pszName)

	: CIThread()
{
	SYSINFO			sysinfo;

	// allow only one application object
	if (m_pApp)
		Fail("More than one application object");

	// initialize member variables
	m_pApp			= this;
	m_bCreated		= FALSE;

	m_main			=
	m_hRTObject		= ::GetRtThreadHandles(THIS_THREAD);
	m_bOwner		= TRUE;

	// parse the name to check the format (no way to pass error back)
	ParseName(pszName, &m_hCatProcess, m_cName);

	// get kernel tick length in usecs
	if (!CopyRtSystemInfo(&sysinfo))
		Fail("Cannot copy system info");
	m_dwKtickInUsecs= 10000 / sysinfo.KernelTickRatio;
	if (m_dwKtickInUsecs == 0)
		Fail("Invalid kernel tick length");
}


///////////////////////////////////////////////////////////////////////
// destructor

CIApp::				~CIApp()
{
}


///////////////////////////////////////////////////////////////////////
// Create (may be called once only)
// returns: 1 if Init fails, otherwise result of Exit

DWORD CIApp::		Create(
	int					argc,
	char **				argv)
{
	// allow only one call to this function
	if (m_bCreated)
		Fail("Cannot create application more than once");
	m_bCreated		= TRUE;

	// copy arguments to member variables
	m_iArg			= argc;
	m_pArg			= argv;

	// activate the application
	if (InitInstance())
		while (DoForever())
			;

	return ExitInstance();
}


///////////////////////////////////////////////////////////////////////
// InitInstance
// returns: TRUE if successful, FALSE if app should terminate

BOOL CIApp::		InitInstance(void)
{
#ifdef _DEBUG
#ifdef __C99__
	_STD cerr << "Entering InitInstance" << _STD endl;
#else
	_STD cout << "Entering InitInstance" << _STD endl;
#endif
#endif

	// catalog the process handle in the root directory
	if (Catalog())
		Fail("Cannot catalog process handle");

	return TRUE;
}


///////////////////////////////////////////////////////////////////////
// DoForever
// returns: TRUE if loop should continue, FALSE if app should stop

BOOL CIApp::		DoForever(void)
{
	EVENTINFO		ei;

	// wait indefinitely for a message
	if (!RtNotifyEvent(0x0f, WAIT_FOREVER, &ei))
		Fail("Cannot receive notification");

	return OnNotify(&ei);
}


///////////////////////////////////////////////////////////////////////
// ExitInstance
// returns: value to be returned from main

int CIApp::			ExitInstance(void)
{
	// remove the process handle from the root directory
	Uncatalog();

#ifdef _DEBUG
#ifdef __C99__
	_STD cerr << "Application terminated" << _STD endl;
#else
	_STD cout << "Application terminated" << _STD endl;
#endif
#endif
	return 0;
}


///////////////////////////////////////////////////////////////////////
// Delete (protected)

void CIApp::		Delete(void)
{
	// do not delete the thread!
	// let it continue so the app terminates

	// do base class deletion
	CIObject::Delete();
}


///////////////////////////////////////////////////////////////////////
// Lookup (should never be called)
// returns: zero or error code

DWORD CIApp::		Lookup (
	char *				pszName,
	DWORD				dwTimeout)
{
	Fail("CIApp::Lookup should not be called");
	return E_TYPE; // just in case..
}


///////////////////////////////////////////////////////////////////////
// Attach (should never be called)
// returns: zero or error code

DWORD CIApp::		Attach(
	RTHANDLE			h)
{
	Fail("CIApp::Attach should not be called");
	return E_TYPE; // just in case..
}


///////////////////////////////////////////////////////////////////////
// Detach (should never be called)

void CIApp::		Detach(void)
{
	Fail("CIApp::Detach should not be called");
}


///////////////////////////////////////////////////////////////////////
// Catalog (protected)
// returns: zero or error code

DWORD CIApp::		Catalog(void)
{
	RTHANDLE		hProcess, hRoot, hOld;

	// avoid error if no name specified
	if (m_cName[0] == 0)
		return (m_status = E_OK);

	// no process name should be specified (i.e. handle must be NULL)
	if (m_hCatProcess != NULL_RTHANDLE)
		return (m_status = E_PARAM);
	
	// catalog the name in the root directory
	hRoot			= ::GetRtThreadHandles(ROOT_PROCESS);
	hProcess		= ::GetRtThreadHandles(THIS_PROCESS);
	if (::CatalogRtHandle(hRoot, hProcess, m_cName))
		return (m_status = E_OK);

	// if the given name already exists,
	// and the existing name refers to a non-existing object,
	// then remove the existing name
	hOld			= ::LookupRtHandle(hRoot, m_cName, NO_WAIT);
	if ((hOld != BAD_RTHANDLE) &&
		(::GetRtHandleType(hOld) == INVALID_TYPE))
	{
		// this is the case mentioned above:
		// remove the old entry and try again
		if (::UncatalogRtHandle(hRoot, m_cName))
			if (::CatalogRtHandle(hRoot, hProcess, m_cName))
				return (m_status = E_OK);
	}

	// something went wrong
	return (m_status = ::GetLastRtError());
}


///////////////////////////////////////////////////////////////////////
// Uncatalog (protected)
// returns: zero or error code

DWORD CIApp::		Uncatalog(void)
{
	RTHANDLE		hProcess;

	// if we did not catalog, we should not Uncatalog either
	if (m_cName[0] == 0)
		return (m_status = E_OK);

	// we know we cataloged in the root directory
	hProcess		= ::GetRtThreadHandles(ROOT_PROCESS);
	if (::UncatalogRtHandle(hProcess, m_cName))
		return (m_status = E_OK);

	return (m_status = ::GetLastRtError());
}


///////////////////////////////////////////////////////////////////////

BOOL CIApp::		OnNotify(
	LPEVENTINFO			pEI)
{
	return FALSE;
}


#if defined(_DEBUG) || defined(__C99__)
///////////////////////////////////////////////////////////////////////

CIApp * CIApp::		GetApp(void)
{
	return m_pApp;
}
#endif


///////////////////////////////////////////////////////////////////////
// UsecsToKticks

DWORD CIApp::		UsecsToKticks(
	DWORD				dwUsecs)
{
	if (dwUsecs == WAIT_FOREVER)
		return WAIT_FOREVER;

	return (dwUsecs + m_dwKtickInUsecs - 1) / m_dwKtickInUsecs;
}


///////////////////////////////////////////////////////////////////////
// GetPciInteruptLevel
// returns: encoded interrupt level or 0xffff for failure

WORD CIApp::		GetPciInterruptLevel(
	WORD				wPciVendorId,
	WORD				wPciDeviceId)
{
	PCIDEV			dev;

	dev.wVendorId	= wPciVendorId;
	dev.wDeviceId	= wPciDeviceId;
	dev.wDeviceIndex= 0;        // we want the first matching device
	if (!PciFindDevice(&dev))
		return 0xFFFF;           // no device found

	if (dev.byIntLine == 0 || dev.byIntLine == 255)
		return 0xFFFF;             // no IRQ on this device, or unrecognised value
	return (dev.byIntLine < 8) ? (dev.byIntLine*16 + 8) 
        :  ((1 + dev.byIntLine/8) * 16 + (dev.byIntLine % 8));
}
