///////////////////////////////////////////////////////////////////////
// implementation of the CIObject class
// last updated: 14apr2004
// Copyright(C) 2004 Micronet Co. All Rights Reserved.

#include <iostream>
#include "rtpp.h"
#include <stdarg.h>
#include <stdlib.h>


///////////////////////////////////////////////////////////////////////
// constructor (for derived classes only)

CIObject::			CIObject(
	WORD				wType)
{
	// initialize general members
	m_hRTObject		= NULL_RTHANDLE;
	m_status		= E_OK;
	m_bOwner		= FALSE;
	m_wType			= wType;
}


///////////////////////////////////////////////////////////////////////
// destructor (should be overruled)

CIObject::			~CIObject(void)
{
	// detach the RT object
	Detach();
}


///////////////////////////////////////////////////////////////////////
// Attach
// returns: zero or error code

DWORD CIObject::	Attach(
	RTHANDLE			h)
{
	m_status		= E_OK;

	// if the object is attached, detach it
	Detach();

	// is the handle of the right type?
	if (InternalGetType(h) != m_wType)
	{
#ifdef _DEBUG
#ifdef __C99__
		_STD cerr << "Attach wrong type" << _STD endl;
#else
		_STD cout << "Attach wrong type" << _STD endl;
#endif
#endif
		m_status	= E_TYPE;
	}
	else
	{
		// all checks OK: attach it
		InternalAttach((char*)NULL, h, FALSE);
	}

	return m_status;
}

		
///////////////////////////////////////////////////////////////////////
// Detach

void CIObject::		Detach(void)
{
	m_status		= E_OK;

	if (InternalDetach())
	{
		Delete();
		m_hRTObject	= NULL_RTHANDLE;
	}
}


///////////////////////////////////////////////////////////////////////
// Lookup
// returns: zero or error code

DWORD CIObject::	Lookup (
	char *				pszName,
	DWORD				dwTimeout)
{
	RTHANDLE		hObject;

	// if the object is attached, detach it
	Detach();

	// get the process handle, and the object name
	if (!ParseName(pszName, &m_hCatProcess, m_cName))
		return m_status;

	// lookup the name in the proper directory
	hObject		= ::LookupRtHandle(m_hCatProcess, m_cName, dwTimeout);
	if (hObject == BAD_RTHANDLE)
		return (m_status = ::GetLastRtError());

	// the name was found: is it the right type?
	if (InternalGetType(hObject) != m_wType)
	{
#ifdef _DEBUG
#ifdef __C99__
		_STD cerr << "Lookup wrong type" << _STD endl;
#else
		_STD cout << "Lookup wrong type" << _STD endl;
#endif
#endif
		return (m_status = E_TYPE);
	}

	// all checks OK: attach it
	InternalAttach((char*)NULL, hObject, FALSE);

	// return success indication
	return (m_status = E_OK);
}


///////////////////////////////////////////////////////////////////////
// Fail (static)

void CIObject::		Fail(
	char *				pszMessage, ...)
{
	va_list		ap;

	va_start(ap, pszMessage);
	vprintf(pszMessage, ap);
	va_end(ap);
	printf("\nError nr=%x\n", GetLastRtError());

	if 	(CIApp::IsMain())
	{
		// this is the mein thread
		exit(0);
	}
	else
	{
		CIDataMbx	delmbx;
		if (delmbx.Lookup("R?EXIT_MBOX", 5000))
			exit(0);
		DWORD		dwTerminate;
		dwTerminate	= TERMINATE;
		delmbx.Send(&dwTerminate, 4);
		SuspendRtThread(NULL_RTHANDLE);
		while (1)
			RtSleep(655349);
	}

	exit(0);
}

	
///////////////////////////////////////////////////////////////////////
// assignment

CIObject & CIObject::
					operator = (
	CIObject &			src)
{
	if (this != &src)	// beware of self-assignment
	{
		// detach the destination from its element
		Detach();

		// source and destination must be of the same type
		if (m_wType != src.m_wType)
			m_status= E_TYPE;
		else
		{
			// if there is an RT object, attach destination to it
			// (fields are copied by InternalAttach)
			if (src.m_hRTObject != NULL_RTHANDLE)
				InternalAttach((char*)NULL, src.m_hRTObject, FALSE);

			m_status= E_OK;
		}
	}
	return *this;
}


///////////////////////////////////////////////////////////////////////
// compare equal

BOOL CIObject::		operator == (
	CIObject &			op2)
{
	// different types always return unequal
	if (m_wType != op2.m_wType)
		return FALSE;

	// two objects are equal if they are attached to the same RT object
	if ((m_hRTObject == op2.m_hRTObject) &&
			(m_hRTObject != NULL_RTHANDLE))
		return TRUE;
	return FALSE;
}


///////////////////////////////////////////////////////////////////////
// compare unequal

BOOL CIObject::		operator != (
	CIObject &			op2)
{
	// different types always return unequal
	if (m_wType != op2.m_wType)
		return TRUE;

	// two objects are equal if they are attached to the same RT object
	if ((m_hRTObject == op2.m_hRTObject) &&
			(m_hRTObject != NULL_RTHANDLE))
		return FALSE;
	return TRUE;
}


///////////////////////////////////////////////////////////////////////
// delete RT object (protected)
// this function should be overruled

void CIObject::		Delete(void)
{
	// the type dependent deletion should be done in the derived class
	m_bOwner		= FALSE;
	m_hRTObject		= NULL_RTHANDLE;
}


///////////////////////////////////////////////////////////////////////
// remove RT object (protected)
// this function may be overruled

void CIObject::		Remove(void)
{
	// remove the RT object without deleting it
}


///////////////////////////////////////////////////////////////////////
// Parse Name (protected)
// returns: TRUE if the format is OK and the process is found,
//          otherewise FALSE

BOOL CIObject::		ParseName(
	char *				pszFullName,
	RTHANDLE *			phProcess,
	char *				pszName)
{
	// max: 2 * name of 12 chars, + 2 * backslash, + zero
	char			szCopy[27];
	char *			pszProcess;
	char *			pszObject;
	int				iLen;

	m_status		= E_PARAM;
	*phProcess		= NULL_RTHANDLE;
	pszName[0]		= 0;

    if( !pszFullName )
        return TRUE;

	if (pszFullName[0] == 0)
		return TRUE;

	// if name is too long, do not accept
	if (::strlen(pszFullName) >= sizeof(szCopy))
		return FALSE;
	::strcpy(szCopy, pszFullName);

	// the name may begin with a backslash
	if (szCopy[0] == '\\')
	{
		// it begins with a backslash:
		// must be followed by a process name and another backslash
		pszProcess	= &szCopy[1];
		pszObject	= ::strchr(pszProcess, '\\');
		if (!pszObject)
			return FALSE;

		// change backslash into zero, so process name is now
		// a proper string
		*pszObject	= 0;
		pszObject++;

		// the process name may not be empty, and it may not exceed 12 chars
		iLen		= ::strlen(pszProcess);
		if ((iLen < 1) || (iLen > 12))
			return FALSE;

		// lookup the process -- no wait
		// if the name is not found, or is not a process, fail
		*phProcess	= ::LookupRtHandle(
						GetRtThreadHandles(ROOT_PROCESS),
						pszProcess,
						NO_WAIT);
		if (*phProcess == BAD_RTHANDLE)
		{
			m_status= E_TIME;
			return FALSE;
		}

		if (::GetRtHandleType(*phProcess) != PROCESS_TYPE)
		{
			m_status= E_TYPE;
			return FALSE;
		}
	}
	else
	{
		// the name does not begin with a backslash:
		// use the current process's directory
		*phProcess	= NULL_RTHANDLE;
		pszObject	= &szCopy[0];
	}

	// the object name may not be empty and it may not exceed 12 chars
	iLen		= ::strlen(pszObject);
	if ((iLen < 1) || (iLen > 12))
		return FALSE;

	// there may be no more backslashes
	if (::strchr(pszObject, '\\') != NULL)
		return FALSE;

	::strcpy(pszName, pszObject);
	m_status		= E_OK;
	return TRUE;
}


///////////////////////////////////////////////////////////////////////
// Catalog (protected)
// returns: zero or error code

DWORD CIObject::	Catalog(void)
{
	RTHANDLE		hOld;

	ASSERT(m_hRTObject != NULL_RTHANDLE);

	// if name is empty: nothing to catalog
	if (m_cName[0] == 0)
		return (m_status = E_OK);

	// catalog the name in the proper directory
	if (::CatalogRtHandle(m_hCatProcess, m_hRTObject, m_cName))
		return (m_status = E_OK);

	// if the given name already exists,
	// and the existing name refers to a non-existing object,
	// then remove the existing name
	hOld			= ::LookupRtHandle(m_hCatProcess, m_cName, NO_WAIT);
	if ((hOld != BAD_RTHANDLE) &&
		(::GetRtHandleType(hOld) == INVALID_TYPE))
	{
		// this is the case mentioned above: remove the old entry and try again
		if (::UncatalogRtHandle(m_hCatProcess, m_cName))
			if (::CatalogRtHandle(m_hCatProcess, m_hRTObject, m_cName))
				return (m_status = E_OK);
	}
	return (m_status = ::GetLastRtError());
}


///////////////////////////////////////////////////////////////////////
// Uncatalog (protected)
// returns: zero or error code

DWORD CIObject::	Uncatalog(void)
{
	ASSERT(m_hRTObject != NULL_RTHANDLE);

	// if name is empty: nothing to uncatalog
	if (m_cName[0] == 0)
		return (m_status = E_OK);

	if (::UncatalogRtHandle(m_hCatProcess, m_cName))
		return (m_status = E_OK);

	return (m_status = ::GetLastRtError());
}

		
///////////////////////////////////////////////////////////////////////
// InternalAttach (protected)
// attach a RT object to an object

void CIObject::		InternalAttach(
	char *				pszName,
	RTHANDLE			h,
	BOOL				bOwner)
{
	ASSERT(m_hRTObject == NULL_RTHANDLE);
	ASSERT(h != NULL_RTHANDLE);

	m_hRTObject		= h;
	m_bOwner		= bOwner;

	if (bOwner)
	{
		// the element now exists; catalog it
		// in the current process, with a proper name
		if (!ParseName(pszName, &m_hCatProcess, m_cName) ||
				(m_hCatProcess != NULL_RTHANDLE))
			m_status= E_PARAM;
		else
			Catalog();
	}
}

		
///////////////////////////////////////////////////////////////////////
// InternalDetach
// returns: TRUE if the RTobject should be deleted, FALSE otherwise

BOOL CIObject::		InternalDetach(void)
{
	// if not attached, no action
	if (m_hRTObject == NULL_RTHANDLE)
		return FALSE;

	// the object is attached to an RT object
	if (m_bOwner)
	{
		Uncatalog();

		// indicate it should be deleted
		if (CIApp::GetApp() != this)
    		return TRUE;
	}

	Remove();

	m_hRTObject		= NULL_RTHANDLE;
	return FALSE;
}

		
///////////////////////////////////////////////////////////////////////
// InternalGetType
// returns: internal type value for the object, or CITYPE_INVALID

WORD CIObject::		InternalGetType(
	RTHANDLE			h)
{
	WORD			wType, wSubType;

	wType			= ::GetRtHandleTypeEx(h, &wSubType);

	switch (wType)
	{
	case THREAD_TYPE:
		return CITYPE_THREAD;

	case MAILBOX_TYPE:
		if (OBJECT_MAILBOX_SUBTYPE == wSubType)
			return CITYPE_OBJECTMBX;
		return CITYPE_DATAMBX;

	case SEMAPHORE_TYPE:
		return CITYPE_SEMAPHORE;

	case REGION_TYPE:
		return CITYPE_REGION;

	case SEGMENT_TYPE:
		return CITYPE_MEMORY;

	case PORT_TYPE:
		return CITYPE_PORT;

	default:
		return CITYPE_INVALID;
	}
}


#if defined(_DEBUG) || defined(__C99__)
///////////////////////////////////////////////////////////////////////
// these functions are only inline for a release, pre 4.0 build

DWORD CIObject::	GetLastError(void) const
{
	return m_status;
}

RTHANDLE CIObject::	GetHandle(void) const
{
	return m_hRTObject;
}

WORD CIObject::		GetType(void) const
{
	return m_wType;
}
#endif
